import sys
import shlex
from subprocess import run
from subprocess import check_output
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory
from pathlib import Path
from shutil import copytree
from shutil import make_archive
from shutil import rmtree

import click
from click_default_group import DefaultGroup
from livereload import Server
from watchdog.observers import Observer

from .convert import Handler
from .convert import generate
from .config import load_config


def shell(command):
    """
    Execute a shell command and return the output as a list of lines.
    """
    return check_output(shlex.split(command)).decode('utf').splitlines()


@click.group(cls=DefaultGroup, default='show')
@click.version_option(
    prog_name='Markdownreveal',
    message='%(prog)s %(version)s'
)
def cli():
    pass


@cli.command()
@click.argument('markdown_file')
@click.option('-h', '--host', type=str, default='localhost',
              help='Listen on IP (default: localhost).')
@click.option('-p', '--port', type=int, default=8123,
              help='Listen on port (default: 8123).')
def show(markdown_file: Path, host: str='localhost', port: int=8123):
    """
    Visualize your presentation (default).
    """
    markdown_file = Path(markdown_file)

    observer = Observer()
    handler = Handler(markdown_file)
    # Initial generation
    generate(markdown_file)
    observer.schedule(handler, '.', recursive=True)
    observer.start()

    server = Server()
    config = load_config()
    server.watch(str(config['output_path'] / '.reload'), delay=0)
    server.serve(
        root=str(config['output_path']),
        restart_delay=0,
        debug=True,
        open_url=True,
        open_url_delay=0,
        host=host,
        port=port,
    )


@cli.command()
@click.argument('markdown_file')
@click.option('-r', '--remote', type=str, default='origin',
              help='Choose a specific remote.')
def upload(markdown_file: Path, remote: str='origin'):
    """
    Upload your presentation.
    """
    markdown_file = Path(markdown_file)

    try:
        shell('git status')
        remote_url = shell('git remote get-url --push ' + remote)[0]
    except CalledProcessError:
        return

    if 'github' not in remote_url:
        error = 'Uploading only supported for GitHub repositories!'
        sys.stderr.write(error + '\n')
        return

    # We copy the directory because `git` does not follow symlinks...
    with TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir) / 'out'
        generate(markdown_file)
        config = load_config()
        copytree(src=str(config['output_path']), dst=str(tmpdir))

        worktree = '--work-tree=%s' % tmpdir
        current = shell('git rev-parse --abbrev-ref HEAD')[0]
        shell('git checkout -B gh-pages')
        shell('git %s add .' % worktree)
        shell('git %s commit -m "Markdownreveal live presentation"' % worktree)
        shell('git %s push -f -u %s gh-pages' % (worktree, remote))
        shell('git %s checkout %s' % (worktree, current))

    repo = Path(remote_url.split(':')[-1])
    url = 'https://%s.github.io/%s/' % (repo.parent, repo.stem)
    sys.stdout.write('Presentation uploaded to:\n\n' + url + '\n\n')


@cli.command()
@click.argument('markdown_file')
def zip(markdown_file: Path):
    """
    Generate a ZIP file with the presentation.
    """
    markdown_file = Path(markdown_file)

    # We copy the directory because `make_archive` cannot follow symlinks...
    with TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir) / 'out'
        generate(markdown_file)
        config = load_config()
        copytree(src=str(config['output_path']), dst=str(tmpdir))
        make_archive(markdown_file.stem, format='zip', root_dir=str(tmpdir))


@cli.command()
@click.argument('markdown_file')
def pdf(markdown_file: str):
    """
    Generate a PDF file with the presentation.
    """
    if markdown_file.startswith('http'):
        presentation = markdown_file
    else:
        markdown_file = Path(markdown_file)
        generate(markdown_file)
        config = load_config()
        presentation = config['output_path'] / 'index.html'

    name = 'slides.pdf'
    command = 'decktape reveal {presentation} {name}'.format(
        presentation=presentation,
        name=name,
    )
    run(shlex.split(command))


@cli.command()
def clean():
    """
    Clean local Markdownreveal files.
    """
    config = load_config()
    rmtree(str(config['local_path']), ignore_errors=True)



import os
import tarfile
from hashlib import sha1
from pathlib import Path
from urllib.request import urlretrieve

import requests

from .typing import Config
from .typing import TarMembers


def latest_project_release(github: str) -> str:
    """
    Fetch the latest project release tag.

    Parameters
    ----------
    github
        The name of the GitHub project.

    Notes
    -----
    For now, only GitHub projects are supported.
    """
    response = requests.get(
        'https://github.com/%s/releases/latest' % github,
        allow_redirects=True,
    )
    return response.url.split('/')[-1]


def clean_tar_members(members: TarMembers) -> TarMembers:
    """
    Strip .tar components (i.e.: remove top-level directory) from members.

    Parameters
    ----------
    members
        A list of .tar members.

    Returns
    -------
        A clean .tar member list with the stripped components.
    """
    clean = []
    for member in members:
        path = Path(member.name)
        if path.is_absolute():
            raise NotImplementedError('Please, report this unexpected error!')
        parts = path.parts[1:]
        if not parts:
            continue
        if parts[0] == 'test':
            continue
        member.name = str(Path(*parts))
        clean.append(member)
    return clean


def initialize_localdir_project(github: str, outdir: Path, localdir: Path,
                                project_version: str, name: str,
                                download_url: str) -> Path:
    """
    Initialize local directory with the specified project.

    Parameters
    ----------
    github
        The name of the GitHub project.
    outdir
        Path where output files will be generated, with a symbolic link to
        the corresponding project downloaded files.
    localdir
        Path to store local files.
    project_version
        String with the project version to use (i.e.: `3.0.1`). The value
        `latest` is also allowed.
    name
        A name for the local downloaded project.
    download_url
        URL to download the project from. In example:
        `'https://github.com/{project}/archive/{version}.tar.gz'`

    Notes
    -----
    For now, only GitHub projects are supported.
    """
    # Download project
    project_path = localdir / name / project_version
    if not project_path.exists():
        if project_version == 'latest':
            project_version = latest_project_release(github=github)
        download_url = download_url.format(project=github,
                                           version=project_version)
        reveal_tar, headers = urlretrieve(download_url)
        with tarfile.open(reveal_tar) as tar:
            members = clean_tar_members(tar.getmembers())
            tar.extractall(str(project_path), members)
        os.remove(reveal_tar)
    symlink = outdir / name
    if symlink.exists():
        symlink.unlink()
    symlink.symlink_to(project_path, target_is_directory=True)


def initialize_localdir_style(outdir: Path, localdir: Path,
                              style_url: str) -> Path:
    """
    Initialize local directory with the required style files.

    Parameters
    ----------
    outdir
        Path where output files will be generated, with a symbolic link to
        the corresponding reveal.js downloaded files.
    localdir
        Path to store local files.
    style_url
        String with the URL to download the style from.
    """
    # Style
    symlink = outdir / 'markdownrevealstyle'
    if symlink.exists():
        symlink.unlink()
    if not style_url:
        return outdir
    style_version = sha1(style_url.encode('utf')).hexdigest()
    style_path = localdir / style_version
    if not style_path.exists():
        style_tar, headers = urlretrieve(style_url)
        with tarfile.open(style_tar) as tar:
            members = clean_tar_members(tar.getmembers())
            tar.extractall(str(style_path), members)
        os.remove(style_tar)
    symlink.symlink_to(style_path, target_is_directory=True)


def initialize_localdir(config: Config) -> Path:
    """
    Initialize local directory with the required reveal.js files.

    Parameters
    ----------
    config
        Markdownreveal configuration.

    Returns
    -------
        Path where output files will be generated, with a symbolic link to
        the corresponding reveal.js downloaded files.
    """
    localdir = config['local_path']

    # Initialize local directory
    outdir = localdir / 'out'
    outdir.mkdir(parents=True, exist_ok=True)

    # reveal.js
    initialize_localdir_project(
        github='hakimel/reveal.js',
        outdir=outdir,
        localdir=localdir,
        project_version=config['reveal_version'],
        name='revealjs',
        download_url='https://github.com/{project}/archive/{version}.tar.gz'
    )

    # KaTeX
    initialize_localdir_project(
        github='Khan/KaTeX',
        outdir=outdir,
        localdir=localdir,
        project_version=config['katex_version'],
        name='katex',
        download_url='https://github.com/{project}/' +
                     'releases/download/{version}/katex.tar.gz'
    )

    # Style
    initialize_localdir_style(outdir, localdir, config['style'])

    return outdir



import re
from typing import List


def find_indexes(haystack: List[str], regex: str) -> List[int]:
    """
    Find indexes in a list where a regular expression matches.

    Parameters
    ----------
    haystack
        List of strings.
    regex
        Regular expression to match.

    Returns
    -------
        The indexes where the regular expression was found.
    """
    return [i for i, item in enumerate(haystack) if re.search(regex, item)]


def find_style_file(filename, config):
    """
    TODO
    """
    outpath = config['local_path'] / 'out'
    filepath = outpath / config['style_path'] / config[filename]
    if not filepath.exists():
        filepath = outpath / 'markdownrevealstyle' / config[filename]
    if not filepath.exists():
        return
    return filepath.relative_to(filepath.parents[1])


def tweak_html_footer(html, footer):
    """
    TODO
    """
    if not footer:
        return False
    text = '<div class="markdownreveal_footer">%s</div>' % footer
    for index in find_indexes(html, '<div class=\"reveal\">'):
        html.insert(index + 1, text)
    return True


def tweak_html_header(html, header):
    """
    TODO
    """
    if not header:
        return
    text = '<div class="markdownreveal_header">%s</div>' % header
    for index in find_indexes(html, '<div class=\"reveal\">'):
        html.insert(index + 1, text)


def tweak_html_warmup(html, config):
    """
    TODO
    """
    fname = find_style_file('style_warmup', config)
    if not fname:
        return
    text = '<section><img src="%s" /></section>' % fname
    index = find_indexes(html, 'div class="slides"')[0]
    html.insert(index + 1, text)


def tweak_html_logo(html, config):
    """
    TODO
    """
    fname = find_style_file('style_logo', config)
    if not fname:
        return
    text = '<div class="logo"><img src="%s" /></div>' % fname
    for index in find_indexes(html, '<div class=\"reveal\">'):
        html.insert(index + 1, text)


def tweak_html_background(html, config):
    """
    TODO
    """
    fname = find_style_file('style_background', config)
    if not fname:
        return
    for index in find_indexes(html, '^<section.*'):
        html[index] = html[index].replace(
            '<section',
            '<section data-background="%s"' % fname,
            1,
        )


def tweak_html_css(html, config):
    """
    TODO
    """
    fname = find_style_file('style_custom_css', config)
    if not fname:
        return
    index = find_indexes(html, 'stylesheet.*id="theme"')[0]
    text = '<link rel="stylesheet" href="%s">' % fname
    html.insert(index + 1, text)
    return True


def tweak_html_emoji(html):
    """
    Add required scripts to parse emojis and display them with a consistent
    style in all browsers.
    """
    text = '''
<script src="https://twemoji.maxcdn.com/v/latest/twemoji.min.js"
  crossorigin="anonymous"></script>
<script>
  function addEvent(element, eventName, fn) {
    if (element.addEventListener)
      element.addEventListener(eventName, fn, false);
    else if (element.attachEvent)
      element.attachEvent('on' + eventName, fn);
  }

  addEvent(window, 'load', function() {
    twemoji.parse(document.body, {'folder': 'svg', 'ext': '.svg'});
  });
</script>
'''
    html.insert(find_indexes(html, '</head>')[0], text)


def tweak_html(html, config):
    """
    TODO
    """
    html = html.splitlines()
    tweak_html_footer(html, config['footer'])
    tweak_html_header(html, config['header'])
    tweak_html_warmup(html, config)
    tweak_html_logo(html, config)
    tweak_html_background(html, config)
    tweak_html_css(html, config)
    tweak_html_emoji(html)
    return '\n'.join(html)



from distutils.version import LooseVersion
from subprocess import check_output
from sys import platform
from threading import Timer
from typing import List

from pypandoc import convert_text
from pypandoc import get_pandoc_version
from watchdog.events import FileSystemEventHandler

from .config import load_config
from .local import initialize_localdir
from .tweak import tweak_html
from .typing import Config


if platform == 'linux':
    from watchdog.observers.inotify_buffer import InotifyBuffer
    # Reduce watchdog default buffer delay for faster response times
    InotifyBuffer.delay = 0.1


def pandoc_extra_to_args(config: Config) -> List[str]:
    """
    Transform Pandoc extra configuration options into Pandoc arguments.

    Parameters
    ----------
    config
        Markdownreveal configuration.

    Returns
    -------
        A list with all the Pandoc arguments.
    """
    arguments = []
    for key, value in config['pandoc_extra'].items():
        if isinstance(value, bool):
            if value:
                arguments.append('--' + key)
        else:
            arguments.append('--{}={}'.format(key, value))
    return arguments


def reveal_extra_to_args(config: Config) -> List[str]:
    """
    Transform reveal.js extra configuration options into Pandoc arguments.

    Parameters
    ----------
    config
        Markdownreveal configuration.

    Returns
    -------
        A list with all the Pandoc arguments.
    """
    arguments = []
    for key, value in config['reveal_extra'].items():
        arguments.extend(['-V', '%s=%s' % (key, value)])
    return arguments


def markdown_to_reveal(text: str, config: Config) -> str:
    """
    Transform a Markdown input file to an HTML (reveal.js) output string.

    Parameters
    ----------
    markdown_text
        Markdown text to convert to HTML.
    config
        Markdownreveal configuration.

    Returns
    -------
        The converted string.
    """
    extra_args = [
        '-s',
        '--slide-level=2',
        '-V', 'revealjs-url=revealjs',
    ]
    if config['katex']:
        pandoc_version = get_pandoc_version()
        if LooseVersion(pandoc_version) < LooseVersion('2.0'):
            extra_args.extend([
                '--katex=katex/katex.min.js',
                '--katex-stylesheet=katex/katex.min.css',
            ])
        else:
            extra_args.extend([
                '--katex=katex/',
            ])
    extra_args.extend(pandoc_extra_to_args(config))
    extra_args.extend(reveal_extra_to_args(config))
    input_format = 'markdown'
    if config['emoji_codes']:
        input_format += '+emoji'
    output = convert_text(
        source=text,
        format=input_format,
        to='revealjs',
        extra_args=extra_args,
    )

    # HTML substitution
    output = tweak_html(output, config)

    return output


def generate(markdown_file):
    """
    TODO
    """
    # Reload config
    config = load_config()

    # Initialize localdir
    initialize_localdir(config)

    # rsync
    command = [
        'rsync',
        '--delete',
        '--exclude', 'katex',
        '--exclude', 'revealjs',
        '--exclude', 'markdownrevealstyle',
        '--exclude', '.git',
        '-av',
        '%s/' % markdown_file.resolve().parent,
        '%s/' % config['output_path'],
    ]
    check_output(command)

    # Convert from markdown
    output = markdown_to_reveal(markdown_file.read_text(), config)

    # Write index.html
    index = config['output_path'] / 'index.html'
    index.write_text(output)

    # Reload view
    with open(str(config['output_path'] / '.reload'), 'a') as f:
        f.write('x\n')


class Handler(FileSystemEventHandler):
    def __init__(self, markdown_file, period=0.1):
        self.markdown_file = markdown_file
        self.period = period
        self.timer = None
        self.set_timer()

    def on_any_event(self, event):
        if self.timer.is_alive():
            return
        self.set_timer()
        self.timer.start()

    def set_timer(self):
        self.timer = Timer(self.period, generate, args=(self.markdown_file,))
